/***************************************************************************
 * jtalk -- talk-like Jabber client
 * (C) Nicolas George - 2006
 * contacts.c: contacts database
 ***************************************************************************/

/*
  This file is part of jtalk.

  jtalk is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  jtalk is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with jtalk; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <string.h>
#include <glib.h>

#include "jtalk.h"

static GStringChunk *names;
static GHashTable *contacts_by_jid;
static GHashTable *contacts_by_short_name;

static jtalk_message_t *all_first = NULL, *all_last = NULL;
static size_t all_size = 0;

static char *
string_until(const char *s, char l, int all)
{
    char *e;
    char *r;

    e = strchr(s, l);
    if(e == NULL)
	return(all ? g_strdup(s) : NULL);
    r = g_new(char, e - s + 1);
    memcpy(r, s, e - s);
    r[e - s] = 0;
    return(r);
}

void
jtalk_contact_init(void)
{
    names = g_string_chunk_new(4096);
    contacts_by_jid = g_hash_table_new(g_str_hash, g_str_equal);
    contacts_by_short_name = g_hash_table_new(g_str_hash, g_str_equal);
}

jtalk_contact_t *
jtalk_contact_find(const char *jid, int add)
{
    char *rjid;
    jtalk_contact_t *c;

    rjid = string_until(jid, '/', 1);
    c = g_hash_table_lookup(contacts_by_jid, rjid);
    if(c == NULL && add) {
	c = g_new(jtalk_contact_t, 1);
	c->jid = g_string_chunk_insert_const(names, rjid);
	c->msg_first = c->msg_last = NULL;
	c->log_file_name = jtalk_log_file_name(c);
	c->subscribed_from = c->subscribed_to = 0;
	c->available_res = NULL;
	c->available_ts = jtalk_now();
	c->undelivered_msg = 0;
	g_hash_table_insert(contacts_by_jid, c->jid, c);
    }
    g_free(rjid);
    return(c);
}

static void
jtalk_contact_get_all_iter(gpointer key, gpointer value, gpointer data)
{
    (void)key;
    GList **l = data;

    *l = g_list_prepend(*l, value);
}

GList *
jtalk_contact_get_all(void)
{
    GList *r;

    r = NULL;
    g_hash_table_foreach(contacts_by_jid, jtalk_contact_get_all_iter, &r);
    return(r);
}

void
jtalk_contact_add_message(jtalk_contact_t *c, jtalk_message_t *m)
{
    GList *l;

    l = g_list_alloc();
    l->data = m;
    l->prev = c->msg_last;
    l->next = NULL;
    if(c->msg_last == NULL)
	c->msg_first = l;
    else
	c->msg_last->next = l;
    c->msg_last = l;

    if(m->from == NULL)
	m->to = g_slist_prepend(m->to, c);
}

static void
jtalk_contact_remove_message(jtalk_contact_t *c, jtalk_message_t *m)
{
    GList *mp;

    /* as is, it will always be the first */
    for(mp = c->msg_first; mp != NULL; mp = mp->next)
	if(mp->data == m)
	    break;
    g_assert(mp != NULL);
    if(mp->prev == NULL)
	c->msg_first = mp->next;
    else
	mp->prev->next = mp->next;
    if(mp->next == NULL)
	c->msg_last = mp->prev;
    else
	mp->next->prev = mp->prev;
    g_list_free_1(mp);
}

static void
jtalk_message_delete(jtalk_message_t *m)
{
    GSList *to;

    if(!m->delivered) {
	g_assert(m->from != NULL);
	m->from->undelivered_msg--;
	/* TODO: warn the user */
    }
    all_size -= sizeof(jtalk_message_t) + strlen(m->text) + 1;
    if(m->all_next == NULL)
	all_last = m->all_prev;
    else
	m->all_next->all_prev = m->all_prev;
    if(m->all_prev == NULL)
	all_first = m->all_next;
    else
	m->all_prev->all_next = m->all_next;
    if(m->from != NULL)
	jtalk_contact_remove_message(m->from, m);
    for(to = m->to; to != NULL; to = to->next)
	jtalk_contact_remove_message(to->data, m);
    g_free(m);
}

jtalk_message_t *
jtalk_message_new(const char *jid, const char *text)
{
    jtalk_message_t *m;
    jtalk_contact_t *c;
    size_t l, lm;
    char *r;

    l = strlen(text);
    lm = sizeof(jtalk_message_t) + l + 1;

    all_size += lm;
    while(all_size > jtalk_opt_max_history && all_first != NULL)
	jtalk_message_delete(all_first);

    m = g_malloc(lm);
    strcpy(m->text, text);
    m->timestamp = jtalk_now();
    m->to = NULL;
    m->delivered = 0;
    m->connection = NULL;

    if(jid != NULL) {
	c = jtalk_contact_find(jid, 1);
	m->from = c;
	r = strchr(jid, '/');
	m->resource = r == NULL ? NULL :
	    g_string_chunk_insert_const(names, r + 1);
	jtalk_contact_add_message(c, m);
    } else {
	m->from = NULL;
	m->resource = NULL;
    }

    m->all_prev = all_last;
    m->all_next = NULL;
    if(all_last == NULL)
	all_first = m;
    else
	all_last->all_next = m;
    all_last = m;

    return(m);
}

void
jtalk_contact_set_available(jtalk_contact_t *c, const char *res, int av)
{
    GSList *ap;
    int old_av;
    int value;
    char ts[9];

    old_av = c->available_res != NULL;
    if(av) {
	for(ap = c->available_res; ap != NULL; ap = ap->next)
	    if(strcmp(ap->data, res) == 0)
		break;
	if(ap == NULL)
	    c->available_res = g_slist_prepend(c->available_res,
		g_string_chunk_insert_const(names, res));
	c->available_ts = jtalk_now();
    } else {
	for(ap = c->available_res; ap != NULL; ap = ap->next)
	    if(strcmp(ap->data, res) == 0)
		break;
	if(ap != NULL) {
	    c->available_res = g_slist_delete_link(c->available_res, ap);
	    if(c->available_res == NULL)
		c->available_ts = jtalk_now();
	}
    }
    if(old_av != (c->available_res != NULL)) {
	value = jtalk_sorter_eval_l(2, "type", "presence",
	    "from", c->jid, "present", av ? "yes" : "no", NULL);
	jtalk_print_timestamp(c->available_ts, ts);
	jtalk_tell_user(value, "(%s) %s<%s>%s is %sthere.",
	    ts, jtalk_esc_name, c->jid, jtalk_esc_reset, av ? "" : "not ");
    }
}

static gint
jtalk_contact_compare(gconstpointer a_v, gconstpointer b_v)
{
    const jtalk_contact_t *a = a_v, *b = b_v;

    return(strcmp(a->jid, b->jid));
}

extern GList *
jtalk_contact_sort(GList *contacts)
{
    return(g_list_sort(contacts, jtalk_contact_compare));
}
