/***************************************************************************
 * jtalk -- talk-like Jabber client
 * (C) Nicolas George - 2006
 * jtalkd.c: startup the connection daemon
 ***************************************************************************/

/*
  This file is part of jtalk.

  jtalk is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  jtalk is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with jtalk; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/stat.h>
#include <glib.h>

#include "jtalk.h"

static void *main_loop;
static pid_t pid_to_watch;

static const char *default_init_file = 
    "# This file is read as commands (without the /) at jtalk startup.\n"
    "# You can use to configure accounts and set default options.\n"
    "#set foo bar\n"
    "#account server=jabber.com user=someone pass=xyzzy tls\n";

static gboolean
jtalk_watch_pid(gpointer dummy)
{
    (void)dummy;
    if(kill(pid_to_watch, 0)) {
	if(errno != ESRCH) {
	    perror("kill");
	    exit(1);
	}
	g_main_loop_quit(main_loop);
    }
    return(TRUE);
}

static void
exec_init_file(void)
{
    char *init_file;
    FILE *file;
    char buf[4096];
    size_t l;

    init_file = jtalk_settings_get_file("init");
    if((file = fopen(init_file, "r")) == NULL) {
	perror(init_file);
	g_free(init_file);
	return;
    }
    g_free(init_file);
    buf[0] = '/';
    while(fgets(buf + 1, sizeof(buf) - 1, file) != NULL) {
	l = strlen(buf);
	if(l > 0 && buf[l - 1] == '\n')
	    buf[--l] = 0;
	if(buf[1] == 0 || buf[1] == '#')
	    continue;
	jtalk_command_from_client(NULL, buf);
    }
    fclose(file);
}

static void
check_or_create_dot_jtalk(void)
{
    struct stat st;
    char *base, *socket, *init;
    FILE *i;

    base = jtalk_settings_get_file(NULL);
    if(stat(base, &st) == 0)
	return;
    if(errno != ENOENT) {
	perror("Warning: failed to stat jtalk base directory");
	return;
    }
    fprintf(stderr, "Creating a skeletton .jtalk directory in:\n"
	"  %s\n"
	"You probably want to kill jtalkd now and edit some things.\n",
	base);
    if(mkdir(base, 0700)) /* there are personnal private information */
	perror(base);
    socket = jtalk_settings_get_file("socket");
    if(mkdir(socket, 0777))
	perror(socket);
    g_free(socket);
    init = jtalk_settings_get_file("init");
    if((i = fopen(init, "w")) == NULL) {
	perror(init);
    } else {
	fprintf(i, default_init_file);
	fclose(i);
    }
    g_free(init);
}

static void
acquire_lock(void)
{
    char *lockf;
    int fd;
    struct flock l;

    lockf = jtalk_settings_get_file("lock");
    if((fd = open(lockf, O_RDWR | O_CREAT, 0666)) < 0) {
	perror(lockf);
	exit(1);
    }
    g_free(lockf);
    l.l_type = F_WRLCK;
    l.l_whence = SEEK_SET;
    l.l_start = 0;
    l.l_len = 0;
    if(fcntl(fd, F_SETLK, &l) < 0) {
	if(errno == EAGAIN)
	    fprintf(stderr,
		"Unable to acquire lock: another jtalkd is running.\n");
	else
	    perror("Unable to acquire the lock");
	exit(1);
    }
}

static void
usage(int err)
{
    fprintf(err ? stderr : stdout,
	"Usage: jtalkd [options]\n"
	"\t-k\tkill the running jtalkd\n"
	"\t\ttwice: kill it until its dead\n"
	"\t\tthrice: ... then start normally\n"
	"\t-l\tlist the running jtalkd and exit\n"
	"\t-f\tfork into the background\n"
	"\t-d pid\texit when pid dies\n"
	"\t\tpid=0 => use the parent\n"
	);
    exit(err);
}

static pid_t
jtalk_init_find_other(void)
{
    char *lockf;
    int fd;
    struct flock l;

    lockf = jtalk_settings_get_file("lock");
    if((fd = open(lockf, O_RDONLY)) < 0)
	return(0);
    g_free(lockf);
    l.l_type = F_WRLCK;
    l.l_whence = SEEK_SET;
    l.l_start = 0;
    l.l_len = 0;
    if(fcntl(fd, F_GETLK, &l)) {
	perror("fcntl");
	exit(1);
    }
    close(fd);
    return(l.l_type == F_UNLCK ? 0 : l.l_pid);
}

static void
jtalk_init_list_other(void)
{
    pid_t o;

    o = jtalk_init_find_other();
    if(o == 0) {
	fprintf(stderr, "No jtalkd running.\n");
	exit(1);
    } else {
	printf("%d\n", o);
	exit(0);
    }
}

static void
jtalk_init_kill_other(int level)
{
    pid_t other;
    int signal = SIGTERM;
    time_t timeout, now;

    time(&timeout);
    timeout += 5;
    while((other = jtalk_init_find_other()) > 0) {
	if(kill(other, signal)) {
	    perror("kill");
	    exit(1);
	}
	if(level <= 1)
	    exit(0);
	usleep(100000);
	if(timeout > 0) {
	    time(&now);
	    if(now > timeout) {
		fprintf(stderr, "It won't die, using a bigger weapon.\n");
		signal = SIGKILL;
		timeout = -1;
	    }
	}
    }
    if(level <= 1) {
	fprintf(stderr, "No jtalkd running.\n");
	exit(1);
    }
    if(level < 3)
	exit(0);
}

static void
jtalk_init_daemonize(void)
{
    int nstdin, nstdout;
    char *log_file;
    pid_t child;

    if((nstdin = open("/dev/null", O_RDONLY)) < 0) {
	perror("/dev/null");
	exit(1);
    }
    log_file = jtalk_settings_get_file("log");
    if((nstdout = open(log_file, O_WRONLY | O_APPEND | O_CREAT, 0666)) < 0) {
	perror(log_file);
	exit(1);
    }
    if(nstdin <= 2 || nstdout <= 2) {
	fprintf(stderr, "Where are your standard streams?\n");
	exit(1);
    }
    if((child = fork()) < 0) {
	perror("Unable to fork");
	exit(1);
    }
    if(child == 0) {
	if(setsid() < 0)
	    perror("setsid");
	if(dup2(nstdin, 0) < 0 || dup2(nstdout, 1) < 0 || dup2(nstdout, 2) < 0)
	    perror("dup2");
	close(nstdin);
	close(nstdout);
	chdir(jtalk_settings_get_file(NULL));
    } else {
	time_t timeout;

	time(&timeout);
	timeout += 3;
	while(jtalk_init_find_other() != child) {
	    usleep(100000);
	}
	exit(0);
    }
}

static void
do_options(int argc, char **argv)
{
    int o;
    int kill = 0;
    int list = 0;
    int fork = 0;
    pid_t die = -1, other;

    while((o = getopt(argc, argv, "klfd:h")) != -1) {
	switch(o) {
	    case 'k':
		kill++;
		break;
	    case 'l':
		list++;
		break;
	    case 'f':
		fork++;
		break;
	    case 'd':
		die = strtol(optarg, NULL, 0);
		break;
	    case 'h':
		usage(0);
	    default:
		usage(1);
	}
    }
    argv += optind;
    argc -= optind;
    if(argc > 0)
	usage(1);
    if(list)
	jtalk_init_list_other();
    if(kill)
	jtalk_init_kill_other(kill);
    if(die >= 0) {
	pid_to_watch = die == 0 ? getppid() : die;
	g_timeout_add(1000, jtalk_watch_pid, NULL);
    }
    if(fork) {
	if((other = jtalk_init_find_other()) != 0) {
	    fprintf(stderr,
		"Another jtalk is running (pid %d), not starting.\n", other);
	    exit(1);
	}
	jtalk_init_daemonize();
    }
}

int
main(int argc, char **argv)
{
    jtalk_settings_init();
    do_options(argc, argv);
    check_or_create_dot_jtalk();
    acquire_lock();
    exec_init_file();
    socket_server_start();
    jtalk_contact_init();
    main_loop = g_main_loop_new(NULL, FALSE);
    g_main_run(main_loop);
}
