/***************************************************************************
 * jtalk -- talk-like Jabber client
 * (C) Nicolas George - 2006
 * log.c: messages logging
 ***************************************************************************/

/*
  This file is part of jtalk.

  jtalk is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  jtalk is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with jtalk; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <glib.h>

#include "jtalk.h"

char *
jtalk_log_file_name(jtalk_contact_t *c)
{
    GString *s;
    unsigned char *j, *p, *r;
    
    j = c->jid;
    s = g_string_sized_new(strlen(j) * 2);
    for(p = j; *p != 0; p++) {
	if(*p == ',' || *p == '-' || (*p == '.' && p > j) ||
	    (*p >= '0' && *p <= '9') || (*p >= '@' && *p <= 'Z') ||
	    *p == '_' || (*p >= 'a' && *p <= 'z'))
	    g_string_append_c(s, *p);
	else
	    g_string_append_printf(s, "%%%02X", *p);
    }
    r = jtalk_settings_get_file(s->str);
    g_string_free(s, TRUE);
    return(r);
}

static void
jtalk_log_append_escaped(GString *s, const char *t, int q)
{
    const char *p;

    while(*t != 0) {
	for(p = t; *p != 0 && *p != '<' && *p != '>' && *p != '&' &&
	    (!q || *p != '"'); p++);
	if(p > t) {
	    g_string_append_len(s, t, p - t);
	    t = p;
	} else {
	    if(*t == '<')
		g_string_append(s, "&lt;");
	    else if(*t == '>')
		g_string_append(s, "&gt;");
	    else if(*t == '&')
		g_string_append(s, "&amp;");
	    else if(*t == '"')
		g_string_append(s, "&quot;");
	    t++;
	}
    }
}

static void
jtalk_log_append_contact(GString *s, const char *a, jtalk_contact_t *c,
    const char *r)
{
    g_string_append_printf(s, " %s=\"", a);
    jtalk_log_append_escaped(s, c->jid, 1);
    if(r != NULL) {
	g_string_append_c(s, '/');
	jtalk_log_append_escaped(s, r, 1);
    }
    g_string_append_c(s, '"');
}

/* TODO: keep more often used files open */
static void
jtalk_log_message_to(char *t, size_t l, jtalk_contact_t *c)
{
    int fd;
    ssize_t w;
    int value;

    if((fd = open(c->log_file_name, O_WRONLY | O_CREAT | O_APPEND, 0666)) < 0) {
	value = jtalk_sorter_eval_l(5, "type", "system", NULL);
	jtalk_tell_user(value, "unable to open %s: %s\n", c->log_file_name,
	    g_strerror(errno));
	return;
    }
    while(l > 0) {
	if((w = write(fd, t, l)) < 0){
	    if(errno == EINTR)
		continue;
	    value = jtalk_sorter_eval_l(5, "type", "system", NULL);
	    jtalk_tell_user(value, "unable to write to %s: %s\n",
		c->log_file_name, g_strerror(errno));
	    break;
	}
	l -= w;
	t += w;
    }
    close(fd);
}

void
jtalk_log_message(jtalk_message_t *m)
{
    time_t ts;
    struct tm *t;
    GString *s;
    GSList *to;

    s = g_string_new(NULL);
    ts = m->timestamp / 1000;
    if((t = gmtime(&ts)) == NULL) {
	perror("localtime");
	exit(1);
    }
    g_string_append_printf(s, "<msg t=\"%04d-%02d-%02dT%02d:%02d:%02dZ\"",
	t->tm_year + 1900, t->tm_mon + 1, t->tm_mday,
	t->tm_hour, t->tm_min, t->tm_sec);
    if(m->from != NULL)
	jtalk_log_append_contact(s, "from", m->from, m->resource);
    for(to = m->to; to != NULL; to = to->next)
	jtalk_log_append_contact(s, "to", to->data, NULL);
    g_string_append_c(s, '>');
    jtalk_log_append_escaped(s, m->text, 0);
    g_string_append(s, "</msg>\n");
    if(m->from != NULL)
	jtalk_log_message_to(s->str, s->len, m->from);
    for(to = m->to; to != NULL; to = to->next)
	jtalk_log_message_to(s->str, s->len, to->data);
}
