/***************************************************************************
 * jtalk -- talk-like Jabber client
 * (C) Nicolas George - 2006
 * sorter.c: sort and prioritize messages
 ***************************************************************************/

/*
  This file is part of jtalk.

  jtalk is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  jtalk is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with jtalk; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <assert.h>
#include <pcre.h>
#include <glib.h>

#include "jtalk.h"

typedef struct jtalk_criterion_s jtalk_criterion_t;
struct jtalk_criterion_s {
    union {
	struct {
	    const char *target;
	    pcre *value;
	} match;
	struct {
	    jtalk_criterion_t *left, *right;
	} ops;
    } op;
    char type;
};

typedef struct {
    char *name;
    int value;
    int priority;
    jtalk_criterion_t *criterion;
} jtalk_sorter_t;

static GList *sorters = NULL;
static int sorters_counter = 0;

static void
jtalk_criterion_free(jtalk_criterion_t *crit)
{
    (void)crit;/* XXX */
}

static int
jtalk_criterion_match_direct(jtalk_criterion_t *crit, char *(*t)[2], int nt)
{
    int i, e;
    int ovector[30];

    for(i = 0; i < nt; i++)
	if(strcmp(t[i][0], crit->op.match.target) == 0)
	    break;
    if(i >= nt)
	return(0);
    e = pcre_exec(crit->op.match.value, NULL, t[i][1], strlen(t[i][1]),
	0, 0, ovector, sizeof(ovector) / sizeof(*ovector));
    return(e >= 0);
}

static int
jtalk_criterion_match(jtalk_criterion_t *crit, char *(*t)[2], int nt)
{
    switch(crit->type) {
	case '~':
	    return(jtalk_criterion_match_direct(crit, t, nt));
	case '&':
	    return(jtalk_criterion_match(crit->op.ops.left, t, nt) &&
		jtalk_criterion_match(crit->op.ops.right, t, nt));
	case '|':
	    return(jtalk_criterion_match(crit->op.ops.left, t, nt) ||
		jtalk_criterion_match(crit->op.ops.right, t, nt));
	case '!':
	    return(!jtalk_criterion_match(crit->op.ops.left, t, nt));
	default:
	    assert(0);
    }
}

static jtalk_criterion_t *
jtalk_criterion_parse(char ***cmd, int *ncmd, GError **error)
{
    jtalk_criterion_t *r, *left, *right;
    char type;
    pcre *re;
    const char *pcre_err;
    int pcre_err_ofs;

    if(*ncmd >= 1 &&
	(strcmp(**cmd, "&") == 0 || strcmp(**cmd, "&&") == 0 ||
	 strcmp(**cmd, "|") == 0 || strcmp(**cmd, "||") == 0 ||
	 strcmp(**cmd, "!") == 0)) {
	type = (**cmd)[0];
	(*cmd)++;
	(*ncmd)--;
	if((left = jtalk_criterion_parse(cmd, ncmd, error)) == NULL)
	    return(NULL);
	if(type == '!') {
	    right = NULL;
	} else {
	    if((right = jtalk_criterion_parse(cmd, ncmd, error)) == NULL) {
		jtalk_criterion_free(left);
		return(NULL);
	    }
	}
	r = g_new(jtalk_criterion_t, 1);
	r->type = type;
	r->op.ops.left = left;
	r->op.ops.right = right;
    } else if(*ncmd >= 2) {
	re = pcre_compile((*cmd)[1], PCRE_DOTALL | PCRE_UTF8, &pcre_err, 
	    &pcre_err_ofs, NULL);
	if(re == NULL) {
	    g_set_error(error, 0, 0, "syntax error at %s (%d): %s",
		(*cmd)[1], pcre_err_ofs, pcre_err);
	    return(NULL);
	}
	r = g_new(jtalk_criterion_t, 1);
	r->type = '~';
	r->op.match.target = g_strdup((*cmd)[0]);
	r->op.match.value = re;
	(*cmd) += 2;
	(*ncmd) -= 2;
    } else {
	g_set_error(error, 0, 0, "syntax error near %s",
	    *ncmd == 0 ? "the end" : **cmd);
	return(NULL);
    }
    return(r);
}

static void
jtalk_sorter_list(jtalk_client_t *client)
{
    (void)client;/* XXX */
}

static void
jtalk_sorter_insert(jtalk_sorter_t *n)
{
    GList *sp;
    jtalk_sorter_t *s;

    for(sp = sorters; sp != NULL; sp = sp->next) {
	s = sp->data;
	if(s->priority <= n->priority)
	    break;
    }
    sorters = g_list_insert_before(sorters, sp, n);
}

static void
jtalk_sorter_add(jtalk_client_t *client, char **cmd, int ncmd)
{
    int value = 0, priority = 0;
    char *name = NULL;
    jtalk_criterion_t *crit = NULL;
    jtalk_sorter_t *sorter;
    GError *error = NULL;

    while(ncmd > 0) {
	if(ncmd >= 2 && strcmp(cmd[0], "name") == 0 && cmd[1][0] != '#') {
	    name = cmd[1];
	    cmd += 2;
	    ncmd -= 2;
	} else if(ncmd >= 2 && (strcmp(cmd[0], "val") == 0||
	    strcmp(cmd[0], "value") == 0)) {
	    value = strtol(cmd[1], NULL, 0);
	    cmd += 2;
	    ncmd -= 2;
	} else if(ncmd >= 2 && strcmp(cmd[0], "priority") == 0) {
	    priority = strtol(cmd[1], NULL, 0);
	    cmd += 2;
	    ncmd -= 2;
	} else if(ncmd >= 2 && strcmp(cmd[0], "on") == 0 && crit == NULL) {
	    cmd++;
	    ncmd--;
	    if((crit = jtalk_criterion_parse(&cmd, &ncmd, &error)) == NULL)
		break;
	} else {
	    g_set_error(&error, 0, 0, "syntax error near %s", cmd[0]);
	    break;
	}
    }
    if(error == NULL && value == 0)
	g_set_error(&error, 0, 0, "no value");
    if(error != NULL) {
	if(crit != NULL)
	    jtalk_criterion_free(crit);
	jtalk_client_printf(client, "%sError: /sorter add: %s.%s",
	    jtalk_esc_error, error->message, jtalk_esc_reset);
	g_error_free(error);
	return;
    }
    sorter = g_new(jtalk_sorter_t, 1);
    sorter->name = name == NULL ? g_strdup_printf("#%d", ++sorters_counter) :
	g_strdup(name);
    sorter->value = value;
    sorter->priority = priority;
    sorter->criterion = crit;
    jtalk_sorter_insert(sorter);
}

void
jtalk_sorter_command(jtalk_client_t *client, char **cmd, int ncmd)
{
    if(ncmd == 0 || (ncmd == 1 && strcmp(cmd[0], "list") == 0)) {
	jtalk_sorter_list(client);
    } else if(strcmp(cmd[0], "add") == 0) {
	jtalk_sorter_add(client, cmd + 1, ncmd - 1);
    } else {
	jtalk_client_printf(client, "%sError: %s: unknown /sorter command.%s",
	    jtalk_esc_error, cmd[0], jtalk_esc_reset);
    }
}

int
jtalk_sorter_eval_v(int base, char *(*t)[2], int nt)
{
    GList *sp;
    jtalk_sorter_t *s;
    int prio = 0, matched = 0;

    for(sp = sorters; sp != NULL; sp = sp->next) {
	s = sp->data;
	if(matched && s->priority < prio)
	    break;
	if(!jtalk_criterion_match(s->criterion, t, nt))
	    continue;
	matched++;
	prio = s->priority;
	base += s->value;
    }
    return(base);
}

int
jtalk_sorter_eval_l(int base, ...)
{
    va_list args;
    int i, n = 0, r;

    va_start(args, base);
    while(va_arg(args, char *) != NULL) {
	assert(va_arg(args, char *) != NULL);
	n++;
    }
    va_end(args);
    {
	char *t[n][2];

	va_start(args, base);
	for(i = 0; i < n; i++) {
	    t[i][0] = va_arg(args, char *);
	    t[i][1] = va_arg(args, char *);
	}
	r = jtalk_sorter_eval_v(base, t, n);
    }
    return(r);
}
